@@ -4,6 +4,7 @@
   Copyright 2011 Steve Markgraf <steve@steve-m.de>
   Copyright 2012-2016 Tormod Volden <debian.tormod@gmail.com>
   Copyright 2013-2016 Antonio Borneo <borneo.antonio@gmail.com>
+  Copyright 2021 Renaud Fivet <renaud.fivet@gmail.com>
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -43,7 +44,7 @@
 #include <windows.h>
 #endif
 
-#define VERSION "0.6"
+#define VERSION "0.7"
 
 /* device globals */
 stm32_t		*stm		= NULL;
@@ -296,8 +297,16 @@ int main(int argc, char* argv[]) {
 		fprintf(diag, "Using Parser : %s\n", parser->name);
 
 		/* We may know from the file how much data there is */
-		if (!use_stdinout && !readwrite_len)
-			readwrite_len = parser->size(p_st);
+		if (!use_stdinout) {
+			if (!start_addr)
+				start_addr = parser->base(p_st);
+			if (start_addr)
+				fprintf(diag, "Location     : %#08x\n", start_addr);
+
+			if (!readwrite_len)
+				readwrite_len = parser->size(p_st);
+			fprintf(diag, "Size         : %u\n", readwrite_len);
+		}
 	} else {
 		parser = &PARSER_BINARY;
 		p_st = parser->init();
@@ -550,6 +559,7 @@ int main(int argc, char* argv[]) {
 			uint32_t left	= end - addr;
 			len		= max_wlen > left ? left : max_wlen;
 			len		= len > size - offset ? size - offset : len;
+			unsigned int reqlen = len ;
 
 			if (parser->read(p_st, buffer, &len) != PARSER_ERR_OK)
 				goto close;
@@ -614,6 +624,8 @@ int main(int argc, char* argv[]) {
 			);
 			fflush(diag);
 
+			if( len < reqlen)	/* Last read already reached EOF */
+				break ;
 		}
 
 		fprintf(diag,	"Done.\n");
